package db

import (
	"context"
	"fmt"

	"getsturdy.com/api/pkg/snapshots"

	"github.com/jmoiron/sqlx"
	"github.com/lib/pq"
)

type Repository interface {
	Create(*snapshots.Snapshot) error
	LatestInWorkspace(context.Context, string) (*snapshots.Snapshot, error)
	GetByCommitSHA(context.Context, string) (*snapshots.Snapshot, error)
	GetByPreviousSnapshotID(context.Context, snapshots.ID) (*snapshots.Snapshot, error)
	ListByIDs(context.Context, []snapshots.ID) ([]*snapshots.Snapshot, error)
	Get(snapshots.ID) (*snapshots.Snapshot, error)
	Update(*snapshots.Snapshot) error
}

type dbrepo struct {
	db *sqlx.DB
}

func NewRepo(db *sqlx.DB) Repository {
	return &dbrepo{db: db}
}

func (r *dbrepo) Create(snapshot *snapshots.Snapshot) error {
	_, err := r.db.NamedExec(`INSERT INTO snapshots
		(id, created_at, previous_snapshot_id, codebase_id, commit_id, workspace_id , action, diffs_count)
		VALUES(:id, :created_at, :previous_snapshot_id, :codebase_id, :commit_id, :workspace_id, :action, :diffs_count)
    	`, &snapshot)
	if err != nil {
		return fmt.Errorf("failed to insert: %w", err)
	}
	return nil
}

func (r *dbrepo) Get(id snapshots.ID) (*snapshots.Snapshot, error) {
	var res snapshots.Snapshot
	err := r.db.Get(&res, `SELECT id, created_at, previous_snapshot_id, codebase_id, commit_id, workspace_id,  action, diffs_count
		FROM snapshots
		WHERE id=$1
		AND deleted_at IS NULL`, id)
	if err != nil {
		return nil, fmt.Errorf("failed to get snapshot: %w", err)
	}
	return &res, nil
}

func (r *dbrepo) LatestInWorkspace(ctx context.Context, workspaceID string) (*snapshots.Snapshot, error) {
	var res snapshots.Snapshot
	if err := r.db.GetContext(ctx, &res, `SELECT id, created_at, previous_snapshot_id, codebase_id, commit_id, workspace_id,  action, diffs_count
		FROM snapshots
		WHERE workspace_id = $1
		AND deleted_at IS NULL
		ORDER BY created_at DESC
		LIMIT 1
	`, workspaceID); err != nil {
		return nil, fmt.Errorf("failed to get latest in workspace: %w", err)
	}
	return &res, nil
}

func (r *dbrepo) Update(snapshot *snapshots.Snapshot) error {
	_, err := r.db.NamedExec(`UPDATE snapshots
		SET deleted_at = :deleted_at
		WHERE id = :id
    	`, &snapshot)
	if err != nil {
		return fmt.Errorf("failed to update: %w", err)
	}
	return nil
}

// GetByCommitSHA returns snapshots by commit_id
func (r *dbrepo) GetByCommitSHA(ctx context.Context, sha string) (*snapshots.Snapshot, error) {
	var res snapshots.Snapshot
	if err := r.db.GetContext(ctx, &res, `SELECT id, created_at, previous_snapshot_id, codebase_id, commit_id, workspace_id,  action, diffs_count
		FROM snapshots
		WHERE commit_id=$1`, sha); err != nil {
		return nil, fmt.Errorf("failed to get by commit sha: %w", err)
	}
	return &res, nil
}

func (r *dbrepo) ListByIDs(ctx context.Context, ids []snapshots.ID) ([]*snapshots.Snapshot, error) {
	if len(ids) == 0 {
		return nil, nil
	}
	var res []*snapshots.Snapshot
	if err := r.db.SelectContext(ctx, &res, `SELECT id, created_at, previous_snapshot_id, codebase_id, commit_id, workspace_id,  action, diffs_count
		FROM snapshots
		WHERE id = ANY($1)`, pq.Array(ids)); err != nil {
		return nil, fmt.Errorf("failed to list by ids: %w", err)
	}
	return res, nil
}

func (r *dbrepo) GetByPreviousSnapshotID(ctx context.Context, previousSnapshotID snapshots.ID) (*snapshots.Snapshot, error) {
	var res snapshots.Snapshot
	if err := r.db.GetContext(ctx, &res, `SELECT id, created_at, previous_snapshot_id, codebase_id, commit_id, workspace_id,  action, diffs_count
		FROM snapshots
		WHERE previous_snapshot_id = $1
		AND deleted_at is NULL
		ORDER BY created_at DESC
		LIMIT 1
	`, previousSnapshotID); err != nil {
		return nil, fmt.Errorf("failed to get by previous snapshot id: %w", err)
	}
	return &res, nil
}
